##############################################################################
#
# Copyright (c) 2001, 2002 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Test Image Data handling

"""
import unittest

from zope.component import adapter
from zope.component import provideAdapter
from zope.component.testing import PlacelessSetup
from zope.interface import implementer
from zope.traversing.browser.interfaces import IAbsoluteURL

from zope.app.file.browser.image import ImageAdd
from zope.app.file.browser.image import ImageData
from zope.app.file.image import Image


class FakeRequest:
    pass


@adapter(Image, FakeRequest)
@implementer(IAbsoluteURL)
class StubAbsoluteURL:

    def __init__(self, *objects):
        pass

    def __str__(self):
        return '/img'

    __call__ = __str__


class ImageDataTest(PlacelessSetup, unittest.TestCase):

    def testData(self):
        image = Image(b'Data')
        data = ImageData()
        data.context = image
        data.request = None
        self.assertEqual(data(), b'Data')

    def testTag(self):
        provideAdapter(StubAbsoluteURL)
        image = Image()
        fe = ImageData()
        fe.context = image
        fe.request = FakeRequest()

        self.assertEqual(
            fe.tag(),
            '<img src="/img" alt="" height="-1" width="-1" border="0" />')
        self.assertEqual(
            fe.tag(scale=.9, width=10, height=10),
            '<img src="/img" alt="" height="9" width="9" border="0" />')
        self.assertEqual(
            fe.tag(alt="Test Image"),
            '<img src="/img" alt="Test Image"'
            ' height="-1" width="-1" border="0" />')
        self.assertEqual(
            fe.tag(height=100, width=100),
            '<img src="/img" alt="" height="100" width="100" border="0" />')
        self.assertEqual(
            fe.tag(border=1),
            '<img src="/img" alt="" height="-1" width="-1" border="1" />')
        self.assertEqual(
            fe.tag(css_class="Image"),
            '<img src="/img" alt=""'
            ' height="-1" width="-1" border="0" class="Image" />')
        self.assertEqual(
            fe.tag(height=100, width="100", border=1, css_class="Image"),
            '<img src="/img" alt=""'
            ' height="100" width="100" class="Image" border="1" />')


class TestImageAdd(unittest.TestCase):

    def test_update_idempotent(self):
        add = ImageAdd()
        add.update_status = "Hi"
        self.assertEqual(add.update(), "Hi")


def test_suite():
    return unittest.defaultTestLoader.loadTestsFromName(__name__)


if __name__ == '__main__':
    unittest.main()



##############################################################################
#
# Copyright (c) 2004 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Tests for zope.app.file.browser.file.

"""
import doctest

from zope.component.testing import setUp
from zope.component.testing import tearDown


def test_suite():
    return doctest.DocTestSuite(
        "zope.app.file.browser.file",
        setUp=setUp,
        tearDown=tearDown,
        optionflags=(doctest.ELLIPSIS
                     | doctest.NORMALIZE_WHITESPACE))



##############################################################################
#
# Copyright (c) 2001, 2002 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Basic File interfaces.

"""
__docformat__ = 'restructuredtext'

from zope.interface import Interface
from zope.schema import Bytes
from zope.schema import NativeStringLine

from zope.app.file.i18n import ZopeMessageFactory as _


class IFile(Interface):

    contentType = NativeStringLine(
        title=_('Content Type'),
        description=_('The content type identifies the type of data.'),
        default='',
        required=False,
        missing_value=''
    )

    data = Bytes(
        title=_('Data'),
        description=_('The actual content of the object.'),
        default=b'',
        missing_value=b'',
        required=False,
    )

    def getSize():
        """Return the byte-size of the data of the object."""


class IImage(IFile):
    """This interface defines an Image that can be displayed.
    """

    def getImageSize():
        """Return a tuple (x, y) that describes the dimensions of
        the object.
        """



##############################################################################
#
# Copyright (c) 2001, 2002 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""File content component
"""
__docformat__ = 'restructuredtext'

import transaction
import zope.app.publication.interfaces
from persistent import Persistent
from zope.interface import implementer

from zope.app.file import interfaces


# set the size of the chunks
MAXCHUNKSIZE = 1 << 16

try:
    text_type = unicode
except NameError:
    text_type = str


@implementer(zope.app.publication.interfaces.IFileContent,
             interfaces.IFile)
class File(Persistent):
    """A persistent content component storing binary file data

    Let's test the constructor:

    >>> file = File()
    >>> file.contentType
    ''
    >>> file.data == b''
    True

    >>> file = File(b'Foobar')
    >>> file.contentType
    ''
    >>> file.data == b'Foobar'
    True

    >>> file = File(b'Foobar', 'text/plain')
    >>> file.contentType
    'text/plain'
    >>> file.data == b'Foobar'
    True

    >>> file = File(data=b'Foobar', contentType='text/plain')
    >>> file.contentType
    'text/plain'
    >>> file.data == b'Foobar'
    True


    Let's test the mutators:

    >>> file = File()
    >>> file.contentType = 'text/plain'
    >>> file.contentType
    'text/plain'

    >>> file.data = b'Foobar'
    >>> file.data == b'Foobar'
    True

    >>> file.data = None
    Traceback (most recent call last):
    ...
    TypeError: Cannot set None data on a file.


    Let's test large data input:

    >>> file = File()

    Insert as string:

    >>> file.data = b'Foobar'*60000
    >>> file.getSize()
    360000
    >>> file.data == b'Foobar'*60000
    True

    Insert data as FileChunk:

    >>> fc = FileChunk(b'Foobar'*4000)
    >>> file.data = fc
    >>> file.getSize()
    24000
    >>> file.data == b'Foobar'*4000
    True

    Insert data from file object:

    >>> from io import BytesIO
    >>> sio = BytesIO()
    >>> _ = sio.write(b'Foobar'*100000)
    >>> _ = sio.seek(0)
    >>> file.data = sio
    >>> file.getSize() == 600000
    True
    >>> file.data == b'Foobar'*100000
    True


    Last, but not least, verify the interface:

    >>> from zope.interface.verify import verifyClass
    >>> interfaces.IFile.implementedBy(File)
    True
    >>> verifyClass(interfaces.IFile, File)
    True
    """

    _data = None
    _size = 0

    def __init__(self, data=b'', contentType=''):
        self.data = data
        self.contentType = contentType

    def _getData(self):
        if isinstance(self._data, FileChunk):
            return bytes(self._data)
        return self._data

    def _setData(self, data):

        # Handle case when data is a string
        if isinstance(data, text_type):
            data = data.encode('UTF-8')

        if isinstance(data, bytes):
            self._data, self._size = FileChunk(data), len(data)
            return

        # Handle case when data is None
        if data is None:
            raise TypeError('Cannot set None data on a file.')

        # Handle case when data is already a FileChunk
        if isinstance(data, FileChunk):
            size = len(data)
            self._data, self._size = data, size
            return

        # Handle case when data is a file object
        seek = data.seek
        read = data.read

        seek(0, 2)
        size = end = data.tell()

        if size <= 2 * MAXCHUNKSIZE:
            seek(0)
            if size < MAXCHUNKSIZE:
                self._data, self._size = read(size), size
                return
            self._data, self._size = FileChunk(read(size)), size
            return

        # Make sure we have an _p_jar, even if we are a new object, by
        # doing a sub-transaction commit.
        transaction.savepoint(optimistic=True)

        jar = self._p_jar

        if jar is None:
            # Ugh
            seek(0)
            self._data, self._size = FileChunk(read(size)), size
            return

        # Now we're going to build a linked list from back
        # to front to minimize the number of database updates
        # and to allow us to get things out of memory as soon as
        # possible.
        next = None
        while end > 0:
            pos = end - MAXCHUNKSIZE
            if pos < MAXCHUNKSIZE:
                pos = 0  # we always want at least MAXCHUNKSIZE bytes
            seek(pos)
            data = FileChunk(read(end - pos))

            # Woooop Woooop Woooop! This is a trick.
            # We stuff the data directly into our jar to reduce the
            # number of updates necessary.
            jar.add(data)

            # This is needed and has side benefit of getting
            # the thing registered:
            data.next = next

            # Now make it get saved in a sub-transaction!
            transaction.savepoint(optimistic=True)

            # Now make it a ghost to free the memory.  We
            # don't need it anymore!
            data._p_changed = None

            next = data
            end = pos

        self._data, self._size = next, size
        return

    def getSize(self):
        '''See `IFile`'''
        return self._size

    # See IFile.
    data = property(_getData, _setData)


class FileChunk(Persistent):
    """Wrapper for possibly large data"""

    next = None

    def __init__(self, data):
        self._data = data

    def __getitem__(self, i):
        return self._data[i]

    def __len__(self):
        data = bytes(self)
        return len(data)

    def __bytes__(self):
        next = self.next
        if next is None:
            return self._data

        result = [self._data]
        while next is not None:
            self = next
            result.append(self._data)
            next = self.next

        return b''.join(result)

    if str is bytes:
        __str__ = __bytes__
    else:
        def __str__(self):
            return self.__bytes__().decode("iso-8859-1", errors='ignore')


class FileReadFile:
    '''Adapter for file-system style read access.

    >>> file = File()
    >>> content = b"This is some file\\ncontent."
    >>> file.data = content
    >>> file.contentType = "text/plain"
    >>> FileReadFile(file).read() == content
    True
    >>> FileReadFile(file).size() == len(content)
    True
    '''

    def __init__(self, context):
        self.context = context

    def read(self):
        return self.context.data

    def size(self):
        return len(self.context.data)


class FileWriteFile:
    """Adapter for file-system style write access.

    >>> file = File()
    >>> content = b"This is some file\\ncontent."
    >>> FileWriteFile(file).write(content)
    >>> file.data == content
    True
    """

    def __init__(self, context):
        self.context = context

    def write(self, data):
        self.context.data = data



##############################################################################
#
# Copyright (c) 2003 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Test HTTP PUT verb

"""

import unittest

from zope.app.file.testing import AppFileLayer
from zope.app.file.tests import http


class TestPUT(unittest.TestCase):

    layer = AppFileLayer

    def test_put(self):
        # PUT something for the first time
        response = http(b"""PUT /testfile.txt HTTP/1.1
Authorization: Basic globalmgr:globalmgrpw
Content-Length: 20
Content-Type: text/plain

This is just a test.""")

        self.assertEqual(response.getStatus(), 201)
        self.assertEqual(response.getHeader("Location"),
                         "http://localhost/testfile.txt")

        response = http(b"""GET /testfile.txt HTTP/1.1
Authorization: Basic globalmgr:globalmgrpw""")
        self.assertEqual(response.getBody(), b"This is just a test.")

        # now modify it
        response = http(b"""PUT /testfile.txt HTTP/1.1
Authorization: Basic globalmgr:globalmgrpw
Content-Length: 23
Content-Type: text/plain

And now it is modified.""")
        self.assertEqual(response.getStatus(), 200)
        self.assertEqual(response.getBody(), b"")

        response = http(b"""GET /testfile.txt HTTP/1.1
Authorization: Basic globalmgr:globalmgrpw""")
        self.assertEqual(response.getBody(), b"And now it is modified.")


def test_suite():
    return unittest.defaultTestLoader.loadTestsFromName(__name__)



##############################################################################
#
# Copyright (c) 2004 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Test File content component.

"""
import doctest
import unittest
from io import BytesIO

from zope.app.file import file as zfile


class TestFile(unittest.TestCase):

    def test_medium_size_chunk(self):
        # MAXCHUNKSIZE < size <= 2 * MAXCHUNKSIZE
        old_size = zfile.MAXCHUNKSIZE
        try:
            zfile.MAXCHUNKSIZE = 10
            data = BytesIO(b'b' * 15)

            f = zfile.File(data)
            self.assertIsInstance(f._data, zfile.FileChunk)
        finally:
            zfile.MAXCHUNKSIZE = old_size

    def test_large_size_chunk_with_jar(self):
        old_size = zfile.MAXCHUNKSIZE
        try:
            zfile.MAXCHUNKSIZE = 10
            data = BytesIO(b'b' * 30)

            class Jar:
                added = ()

                def add(self, o):
                    self.added += (o,)

                def register(self, o):
                    pass

            f = zfile.File()
            f._p_jar = Jar()

            f.data = data

            self.assertEqual(3, len(f._p_jar.added))

            self.assertEqual(data.getvalue(), f.data)
            self.assertEqual('b' * 30, str(f._data))
        finally:
            zfile.MAXCHUNKSIZE = old_size


class TestFileChunk(unittest.TestCase):

    def test_getitem(self):
        chunk = zfile.FileChunk(b'abc')
        self.assertEqual(chunk[:], b'abc')

    def test_bytes(self):
        chunk = zfile.FileChunk(b'abc')
        self.assertEqual(bytes(chunk), b'abc')
        self.assertEqual('abc', str(chunk))


def test_suite():
    return unittest.TestSuite((
        unittest.defaultTestLoader.loadTestsFromName(__name__),
        doctest.DocTestSuite('zope.app.file.file')
    ))



##############################################################################
#
# Copyright (c) 2001, 2002 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Test Image content component

"""
import unittest

from zope.interface.verify import verifyClass

from zope.app.file.file import File
from zope.app.file.file import FileReadFile
from zope.app.file.file import FileWriteFile
from zope.app.file.image import FileFactory
from zope.app.file.image import Image
from zope.app.file.image import ImageSized
from zope.app.file.interfaces import IImage


zptlogo = (
    b'GIF89a\x10\x00\x10\x00\xd5\x00\x00\xff\xff\xff\xff\xff\xfe\xfc\xfd\xfd'
    b'\xfa\xfb\xfc\xf7\xf9\xfa\xf5\xf8\xf9\xf3\xf6\xf8\xf2\xf5\xf7\xf0\xf4\xf6'
    b'\xeb\xf1\xf3\xe5\xed\xef\xde\xe8\xeb\xdc\xe6\xea\xd9\xe4\xe8\xd7\xe2\xe6'
    b'\xd2\xdf\xe3\xd0\xdd\xe3\xcd\xdc\xe1\xcb\xda\xdf\xc9\xd9\xdf\xc8\xd8\xdd'
    b'\xc6\xd7\xdc\xc4\xd6\xdc\xc3\xd4\xda\xc2\xd3\xd9\xc1\xd3\xd9\xc0\xd2\xd9'
    b'\xbd\xd1\xd8\xbd\xd0\xd7\xbc\xcf\xd7\xbb\xcf\xd6\xbb\xce\xd5\xb9\xcd\xd4'
    b'\xb6\xcc\xd4\xb6\xcb\xd3\xb5\xcb\xd2\xb4\xca\xd1\xb2\xc8\xd0\xb1\xc7\xd0'
    b'\xb0\xc7\xcf\xaf\xc6\xce\xae\xc4\xce\xad\xc4\xcd\xab\xc3\xcc\xa9\xc2\xcb'
    b'\xa8\xc1\xca\xa6\xc0\xc9\xa4\xbe\xc8\xa2\xbd\xc7\xa0\xbb\xc5\x9e\xba\xc4'
    b'\x9b\xbf\xcc\x98\xb6\xc1\x8d\xae\xbaFgs\x00\x00\x00\x00\x00\x00\x00\x00'
    b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
    b'\x00,\x00\x00\x00\x00\x10\x00\x10\x00\x00\x06z@\x80pH,\x12k\xc8$\xd2f'
    b'\x04\xd4\x84\x01\x01\xe1\xf0d\x16\x9f\x80A\x01\x91\xc0ZmL\xb0\xcd\x00V'
    b'\xd4\xc4a\x87z\xed\xb0-\x1a\xb3\xb8\x95\xbdf8\x1e\x11\xca,MoC$\x15\x18{'
    b'\x006}m\x13\x16\x1a\x1f\x83\x85}6\x17\x1b $\x83\x00\x86\x19\x1d!%)\x8c'
    b'\x866#\'+.\x8ca`\x1c`(,/1\x94B5\x19\x1e"&*-024\xacNq\xba\xbb\xb8h\xbeb'
    b'\x00A\x00;'
)


class TestImage(unittest.TestCase):

    def _makeImage(self, *args, **kw):
        return Image(*args, **kw)

    def testEmpty(self):
        file = self._makeImage()
        self.assertEqual(file.contentType, '')
        self.assertEqual(file.data, b'')

    def testConstructor(self):
        file = self._makeImage(b'Data')
        self.assertEqual(file.contentType, '')
        self.assertEqual(file.data, b'Data')

    def testMutators(self):
        image = self._makeImage()

        image.contentType = 'image/jpeg'
        self.assertEqual(image.contentType, 'image/jpeg')

        image._setData(zptlogo)
        self.assertEqual(image.data, zptlogo)
        self.assertEqual(image.contentType, 'image/gif')
        self.assertEqual(image.getImageSize(), (16, 16))

    def testInterface(self):
        self.assertTrue(IImage.implementedBy(Image))
        self.assertTrue(verifyClass(IImage, Image))


class TestFileAdapters(unittest.TestCase):

    def _makeFile(self, *args, **kw):
        return Image(*args, **kw)

    def test_ReadFile(self):
        file = self._makeFile()
        content = b"This is some file\ncontent."
        file.data = content
        file.contentType = 'text/plain'
        self.assertEqual(FileReadFile(file).read(), content)
        self.assertEqual(FileReadFile(file).size(), len(content))

    def test_WriteFile(self):
        file = self._makeFile()
        content = b"This is some file\ncontent."
        FileWriteFile(file).write(content)
        self.assertEqual(file.data, content)


class DummyImage:

    def __init__(self, width, height, bytes):
        self.width = width
        self.height = height
        self.bytes = bytes

    def getSize(self):
        return self.bytes

    def getImageSize(self):
        return self.width, self.height


class TestFileFactory(unittest.TestCase):

    def test_image(self):
        factory = FileFactory(None)
        f = factory("spam.txt", "image/foo", b"hello world")
        self.assertIsInstance(f, Image)
        f = factory("spam.txt", "", zptlogo)
        self.assertIsInstance(f, Image)

    def test_text(self):
        factory = FileFactory(None)
        f = factory("spam.txt", "", b"hello world")
        self.assertIsInstance(f, File)
        self.assertNotIsInstance(f, Image)
        f = factory("spam.txt", "", b"\0\1\2\3\4")
        self.assertIsInstance(f, File)
        self.assertNotIsInstance(f, Image)
        f = factory("spam.txt", "text/splat", zptlogo)
        self.assertIsInstance(f, File)
        self.assertNotIsInstance(f, Image)
        f = factory("spam.txt", "application/splat", zptlogo)
        self.assertIsInstance(f, File)
        self.assertNotIsInstance(f, Image)


class TestSized(unittest.TestCase):

    def testInterface(self):
        from zope.size.interfaces import ISized
        self.assertTrue(ISized.implementedBy(ImageSized))
        self.assertTrue(verifyClass(ISized, ImageSized))

    def test_zeroSized(self):
        s = ImageSized(DummyImage(0, 0, 0))
        self.assertEqual(s.sizeForSorting(), ('byte', 0))
        self.assertEqual(s.sizeForDisplay(), '0 KB ${width}x${height}')
        self.assertEqual(s.sizeForDisplay().mapping['width'], '0')
        self.assertEqual(s.sizeForDisplay().mapping['height'], '0')

    def test_arbitrarySize(self):
        s = ImageSized(DummyImage(34, 56, 78))
        self.assertEqual(s.sizeForSorting(), ('byte', 78))
        self.assertEqual(s.sizeForDisplay(), '1 KB ${width}x${height}')
        self.assertEqual(s.sizeForDisplay().mapping['width'], '34')
        self.assertEqual(s.sizeForDisplay().mapping['height'], '56')

    def test_unknownSize(self):
        s = ImageSized(DummyImage(-1, -1, 23))
        self.assertEqual(s.sizeForSorting(), ('byte', 23))
        self.assertEqual(s.sizeForDisplay(), '1 KB ${width}x${height}')
        self.assertEqual(s.sizeForDisplay().mapping['width'], '?')
        self.assertEqual(s.sizeForDisplay().mapping['height'], '?')


class TestGetImageInfo(unittest.TestCase):

    def _info(self, data):
        from zope.app.file.image import getImageInfo
        return getImageInfo(data)

    def test_getImageInfo(self):
        t, w, h = self._info(
            b"\xff\xd8\xff\xe0\x00\x10JFIF\x00\x01\x01\x00\x00\x01\x00\x01\x00"
            b"\x00\xff\xdb\x00C")
        self.assertEqual(t, "image/jpeg")
        self.assertEqual(w, -1)
        self.assertEqual(h, -1)

    def test_getImageInfo_jpeg(self):
        data = (
            b'\xff\xd8\x00\xff\xe0\x00\x10JFIF\x00\x01\x02\x00\x00\x01\x00\x01'
            b'\x00\x00\xff\xdb\x00C\x00\t\x06\x07\x08\x07\x06\t\x08\x08\x08\n'
            b'\n\t\x0b\x0e\x17\x0f\x0e\r\r\x0e\x1c\x14\x15\x11\x17"\x1e##!\x1e'
            b'  %*5-%\'2(  .?/279<<<$-BFA:F5;<9\xff\xdb\x00C\x01\n\n\n\x0e'
            b'\x0c\x0e\x1b\x0f\x0f\x1b9& &999999999999999999999999999999999999'
            b'99999999999999\xff\xc0\x00\x11\x08\x04\x80\x08\x00')

        t, w, h = self._info(data)
        self.assertEqual(t, "image/jpeg")
        self.assertEqual(w, 2048)
        self.assertEqual(h, 1152)

    def test_getImageInfo_bmp(self):
        t, w, h = self._info(b'BMl\x05\x00\x00\x00\x00\x00\x006\x04\x00\x00('
                             b'\x00\x00\x00\x10\x00\x00\x00\x10\x00\x00\x00'
                             b'\x01\x00\x08\x00\x01\x00\x00\x006\x01\x00\x00'
                             b'\x12\x0b\x00\x00\x12\x0b\x00\x00\x00\x01\x00'
                             b'... and so on ...')
        self.assertEqual(t, "image/x-ms-bmp")
        self.assertEqual(w, 16)
        self.assertEqual(h, 16)

    def test_getImageInfo_png(self):
        data = (
            b'\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR\x00\x00\x05\x1c\x00\x00\x01'
            b'p\x08\x06\x00\x00\x00\xda\x1b\xb7')
        t, w, h = self._info(data)
        self.assertEqual(t, "image/png")
        self.assertEqual(w, 1308)
        self.assertEqual(h, 368)


def test_suite():
    return unittest.defaultTestLoader.loadTestsFromName(__name__)



#
# This file is necessary to make this directory a package.

from zope.app.wsgi.testlayer import http as _http

from zope.app.file.testing import AppFileLayer


def http(query_str, *args, **kwargs):
    wsgi_app = AppFileLayer.make_wsgi_app()
    kwargs.setdefault('handle_errors', False)
    return _http(wsgi_app, query_str, *args, **kwargs)



##############################################################################
#
# Copyright (c) 2001, 2002 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""File views.

"""

from datetime import datetime

import zope.contenttype.parse
import zope.datetime
import zope.event
from zope.contenttype import guess_content_type
from zope.dublincore.interfaces import IDCTimes
from zope.exceptions.interfaces import UserError
from zope.schema import Text

from zope import lifecycleevent
from zope.app.file.file import File
from zope.app.file.i18n import ZopeMessageFactory as _
from zope.app.file.interfaces import IFile


__docformat__ = 'restructuredtext'


class FileView:

    request = None
    context = None

    def show(self):
        """Sets various headers if the request is present and returns the
        data of the file. If the "If-Modified-Since" header is set and
        the context is adaptable to IDCTimes, data is only returned if
        the modification date of the context is newer than the date in the
        "If-Modified-Since" header
        >>> from zope.publisher.browser import BrowserView, TestRequest
        >>> class FileTestView(FileView, BrowserView): pass
        >>> import pytz
        >>> class MyFile(object):
        ...     contentType = 'text/plain'
        ...     data = b"data of file"
        ...     modified = datetime(2006,1,1,tzinfo=pytz.utc)
        ...     def getSize(self):
        ...         return len(self.data)

        >>> aFile = MyFile()
        >>> request = TestRequest()
        >>> view = FileTestView(aFile,request)
        >>> view.show() == MyFile.data
        True
        >>> request.response.getHeader('Content-Type')
        'text/plain'
        >>> request.response.getHeader('Content-Length')
        '12'

        If the file is adaptable to IDCTimes the "Last-Modified" header is also
        set.

        >>> request.response.getHeader('Last-Modified') is None
        True

        For the test we just declare that our file provides
        IZopeDublinCore
        >>> from zope.dublincore.interfaces import IZopeDublinCore
        >>> from zope.interface import directlyProvides
        >>> directlyProvides(aFile,IZopeDublinCore)
        >>> request = TestRequest()
        >>> view = FileTestView(aFile,request)
        >>> view.show() == MyFile.data
        True
        >>> request.response.getHeader('Last-Modified')
        'Sun, 01 Jan 2006 00:00:00 GMT'

        If the "If-Modified-Since" header is set and is newer a 304
        status is returned and the value is empty.

        >>> modified = datetime(2007,12,31,tzinfo=pytz.utc)
        >>> modHeader = zope.datetime.rfc1123_date(
        ...     zope.datetime.time(modified.isoformat()))
        >>> request = TestRequest(IF_MODIFIED_SINCE=modHeader)

        >>> view = FileTestView(aFile,request)
        >>> view.show() == b''
        True
        >>> request.response.getStatus()
        304

        An invalid value for that header is ignored.
        >>> request = TestRequest(IF_MODIFIED_SINCE="Not Valid")

        >>> view = FileTestView(aFile,request)
        >>> view.show() == MyFile.data
        True

        """

        if self.request is not None:
            self.request.response.setHeader('Content-Type',
                                            self.context.contentType)
            self.request.response.setHeader('Content-Length',
                                            self.context.getSize())
        try:
            modified = IDCTimes(self.context).modified
        except TypeError:
            modified = None
        if modified is None or not isinstance(modified, datetime):
            return self.context.data

        header = self.request.getHeader('If-Modified-Since', None)
        lmt = zope.datetime.time(modified.isoformat())
        if header is not None:
            header = header.split(';')[0]
            try:
                mod_since = int(zope.datetime.time(header))
            except zope.datetime.SyntaxError:
                mod_since = None
            if mod_since is not None:
                if lmt <= mod_since:
                    self.request.response.setStatus(304)
                    return b''
        self.request.response.setHeader('Last-Modified',
                                        zope.datetime.rfc1123_date(lmt))

        return self.context.data


def cleanupFileName(filename):
    return filename.split('\\')[-1].split('/')[-1]


class FileUpdateView:

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def errors(self):
        form = self.request.form
        if "UPDATE_SUBMIT" in form:
            filename = getattr(form["field.data"], "filename", None)
            contenttype = form.get("field.contentType")
            if filename:
                filename = cleanupFileName(filename)
                if not contenttype:
                    contenttype = guess_content_type(filename)[0]
                if not form.get("add_input_name"):
                    form["add_input_name"] = filename
            return self.update_object(form["field.data"], contenttype)
        return ''


class FileAdd(FileUpdateView):
    """View that adds a new File object based on a file upload.

    >>> class FauxAdding(object):
    ...     def add(self, content):
    ...         self.content = content
    ...     def nextURL(self):
    ...         return 'next url'

    >>> from zope.publisher.browser import TestRequest
    >>> from io import BytesIO
    >>> sio = BytesIO(b"some data")
    >>> sio.filename = 'abc.txt'

    Let's make sure we can use the uploaded file name if one isn't
    specified by the user, and can use the content type when
    specified.

    >>> request = TestRequest(form={'field.data': sio,
    ...                             'field.contentType': 'text/foobar',
    ...                             'UPDATE_SUBMIT': 'Add'})
    >>> adding = FauxAdding()
    >>> view = FileAdd(adding, request)
    >>> view.errors()
    ''
    >>> adding.content.contentType
    'text/foobar'
    >>> adding.content.data == b'some data'
    True
    >>> request.form['add_input_name']
    'abc.txt'

    Now let's guess the content type, but also use a provided file
    name for adding the new content object:

    >>> request = TestRequest(form={'field.data': sio,
    ...                             'field.contentType': '',
    ...                             'add_input_name': 'splat.txt',
    ...                             'UPDATE_SUBMIT': 'Add'})
    >>> adding = FauxAdding()
    >>> view = FileAdd(adding, request)
    >>> view.errors()
    ''
    >>> adding.content.contentType
    'text/plain'
    >>> request.form['add_input_name']
    'splat.txt'

    """

    def update_object(self, data, contenttype):
        f = File(data, contenttype)
        zope.event.notify(lifecycleevent.ObjectCreatedEvent(f))
        self.context.add(f)
        self.request.response.redirect(self.context.nextURL())
        return ''


class FileUpload(FileUpdateView):
    """View that updates an existing File object with a new upload.
        Fires an ObjectModifiedEvent.


    >>> from zope.publisher.browser import TestRequest
    >>> from io import BytesIO
    >>> sio = BytesIO(b"some data")
    >>> sio.filename = 'abc.txt'

    Before we instanciate the request, we need to make sure that the
    ``IUserPreferredLanguages`` adapter exists, so that the request's
    locale exists.  This is necessary because the ``update_object``
    method uses the locale formatter for the status message:

    >>> from zope import component as ztapi
    >>> from zope.publisher.browser import BrowserLanguages
    >>> from zope.publisher.interfaces.http import IHTTPRequest
    >>> from zope.i18n.interfaces import IUserPreferredLanguages
    >>> ztapi.provideAdapter(
    ...     BrowserLanguages, (IHTTPRequest,), IUserPreferredLanguages)

    We install an event logger so we can see the events generated:

    >>> def eventLog(event):
    ...     print('ModifiedEvent:', event.descriptions[0].attributes)
    >>> zope.event.subscribers.append(eventLog)

    Let's make sure we can use the uploaded file name if one isn't
    specified by the user, and can use the content type when
    specified.


    >>> request = TestRequest(form={'field.data': sio,
    ...                             'field.contentType': 'text/foobar',
    ...                             'UPDATE_SUBMIT': 'Update'})
    >>> file = File()
    >>> view = FileUpload(file, request)
    >>> view.errors()
    ModifiedEvent: ('contentType', 'data')
    'Updated on ${date_time}'
    >>> file.contentType
    'text/foobar'
    >>> file.data == b'some data'
    True

    Now let's guess the content type, but also use a provided file
    name for adding the new content object:

    >>> request = TestRequest(form={'field.data': sio,
    ...                             'field.contentType': '',
    ...                             'add_input_name': 'splat.txt',
    ...                             'UPDATE_SUBMIT': 'Update'})
    >>> file = File()
    >>> view = FileUpload(file, request)
    >>> view.errors()
    ModifiedEvent: ('contentType', 'data')
    'Updated on ${date_time}'
    >>> file.contentType
    'text/plain'

    The ObjectModifiedEvent lists only the contentType if the data
    are omitted:

    >>> request = TestRequest(form={'field.data': None,
    ...                             'field.contentType': '',
    ...                             'add_input_name': 'splat.txt',
    ...                             'UPDATE_SUBMIT': 'Update'})
    >>> file = File()
    >>> view = FileUpload(file, request)
    >>> view.errors()
    ModifiedEvent: ('contentType',)
    'Updated on ${date_time}'


    Cleanup:

    >>> zope.event.subscribers.remove(eventLog)

    """

    def update_object(self, data, contenttype):
        self.context.contentType = contenttype

        descriptor = lifecycleevent.Attributes(IFile, "contentType")

        # Update *only* if a new value is specified
        if data:
            self.context.data = data
            descriptor.attributes += ("data",)

        event = lifecycleevent.ObjectModifiedEvent(self.context, descriptor)
        zope.event.notify(event)

        formatter = self.request.locale.dates.getFormatter(
            'dateTime', 'medium')
        return _("Updated on ${date_time}",
                 mapping={'date_time': formatter.format(datetime.utcnow())})


class IFileEditForm(IFile):
    """Schema for the File edit form.

    Replaces the Bytes `data` field with a Text field.
    """

    data = Text(
        title=_('Data'),
        description=_('The actual content of the object.'),
        default='',
        missing_value='',
        required=False,
    )


class UnknownCharset(Exception):
    """Unknown character set."""


class CharsetTooWeak(Exception):
    """Character set cannot encode all characters in text."""


class FileEdit:
    r"""File edit form mixin.

    Lets the user edit a text file directly via a browser form.

    Converts between Unicode strings used in browser forms and 8-bit strings
    stored internally.

        >>> from zope.publisher.browser import BrowserView, TestRequest
        >>> class FileEditView(FileEdit, BrowserView): pass
        >>> view = FileEditView(File(), TestRequest())
        >>> view.getData()['data']
        ''
        >>> view.getData()['contentType']
        ''

        We install an event logger so we can see the events generated.

        >>> def eventLog(event):
        ...    print(event.__class__.__name__,
        ...          event.descriptions[0].attributes)
        >>> zope.event.subscribers.append(eventLog)

        >>> view.setData({'contentType': 'text/plain; charset=ISO-8859-13',
        ...               'data': 'text \u0105'}) # doctest:+ELLIPSIS
        ObjectModifiedEvent ('data', 'contentType')
        'Updated on ${date_time}'

        >>> view.context.contentType
        'text/plain; charset=ISO-8859-13'
        >>> view.context.data == b'text \xe0'
        True

        >>> view.getData()['data']
        'text \u0105'

        Cleanup eventlog.

        >>> zope.event.subscribers.remove(eventLog)

    You will get an error if you try to specify a charset that cannot encode
    all the characters

        >>> view.setData({'contentType': 'text/xml; charset=ISO-8859-1',
        ...               'data': 'text \u0105'})
        Traceback (most recent call last):
          ...
        zope.app.file.browser.file.CharsetTooWeak: ISO-8859-1

    You will get a different error if you try to specify an invalid charset

        >>> view.setData({'contentType': 'text/xml; charset=UNKNOWN',
        ...               'data': 'text \u0105'})
        Traceback (most recent call last):
          ...
        zope.app.file.browser.file.UnknownCharset: UNKNOWN

    The update method catches those errors and replaces them with error
    messages

        >>> from zope.i18n import translate
        >>> class FakeFormView(BrowserView):
        ...     def update(self):
        ...         raise CharsetTooWeak('ASCII')
        >>> class FileEditView(FileEdit, FakeFormView): pass
        >>> view = FileEditView(File(), TestRequest())
        >>> translate(view.update())
        'The character set you specified (ASCII) cannot encode all characters in text.'
        >>> translate(view.update_status)
        'The character set you specified (ASCII) cannot encode all characters in text.'

        >>> class FakeFormView(BrowserView):
        ...     def update(self):
        ...         raise UnknownCharset('UNKNOWN')
        >>> class FileEditView(FileEdit, FakeFormView): pass
        >>> view = FileEditView(File(), TestRequest())
        >>> translate(view.update())
        'The character set you specified (UNKNOWN) is not supported.'
        >>> translate(view.update_status)
        'The character set you specified (UNKNOWN) is not supported.'

    Speaking about errors, if you trick the system and upload a file with
    incorrect charset designation, you will get a UserError when you visit the
    view:

        >>> view.context.contentType = 'text/plain; charset=UNKNOWN'
        >>> view.context.data = b'\xff'
        >>> view.getData()
        Traceback (most recent call last):
          ...
        zope.exceptions.interfaces.UserError: The character set specified in the content type ($charset) is not supported.

        >>> view.context.contentType = 'text/plain; charset=UTF-8'
        >>> view.context.data = b'\xff'
        >>> view.getData()
        Traceback (most recent call last):
          ...
        zope.exceptions.interfaces.UserError: The character set specified in the content type ($charset) does not match file content.

    """  # noqa: E501 line too long
    context = None
    request = None
    error = None

    def getData(self):
        charset = extractCharset(self.context.contentType)
        try:
            return {'contentType': self.context.contentType,
                    'data': self.context.data.decode(charset)}
        except LookupError:
            msg = _("The character set specified in the content type"
                    " ($charset) is not supported.",
                    mapping={'charset': charset})
            raise UserError(msg)
        except UnicodeDecodeError:
            msg = _("The character set specified in the content type"
                    " ($charset) does not match file content.",
                    mapping={'charset': charset})
            raise UserError(msg)

    def setData(self, data):
        charset = extractCharset(data['contentType'])
        try:
            encodeddata = data['data'].encode(charset)
        except LookupError:
            raise UnknownCharset(charset)
        except UnicodeEncodeError:
            raise CharsetTooWeak(charset)

        modified = []
        if encodeddata != self.context.data:
            self.context.data = encodeddata
            modified.append('data')

        if self.context.contentType != data['contentType']:
            self.context.contentType = data['contentType']
            modified.append('contentType')
        formatter = self.request.locale.dates.getFormatter('dateTime',
                                                           'medium')
        if modified:
            event = lifecycleevent.ObjectModifiedEvent(
                self.context,
                lifecycleevent.Attributes(IFile, *modified))
            zope.event.notify(event)

        return _("Updated on ${date_time}",
                 mapping={'date_time': formatter.format(datetime.utcnow())})

    def update(self):
        try:
            return super().update()
        except CharsetTooWeak as charset:
            self.update_status = _("The character set you specified ($charset)"
                                   " cannot encode all characters in text.",
                                   mapping={'charset': charset})
            return self.update_status
        except UnknownCharset as charset:
            self.update_status = _("The character set you specified ($charset)"
                                   " is not supported.",
                                   mapping={'charset': charset})
            return self.update_status


def extractCharset(content_type):
    """Extract charset information from a MIME type.

        >>> extractCharset('text/plain; charset=US-ASCII')
        'US-ASCII'
        >>> extractCharset('text/html; charset=ISO-8859-1')
        'ISO-8859-1'
        >>> extractCharset('text/plain')
        'UTF-8'

    """
    if content_type and content_type.strip():
        _major, _minor, params = zope.contenttype.parse.parse(content_type)
        return params.get("charset", "UTF-8")

    return "UTF-8"



##############################################################################
#
# Copyright (c) 2003, 2004 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Functional tests for File and Image.

"""
import doctest
import unittest
from io import BytesIO
from xml.sax.saxutils import escape

from webtest import TestApp
from zope.app.wsgi.testlayer import encodeMultipartFormdata
from zope.interface.interfaces import ComponentLookupError

from zope.app.file.file import File
from zope.app.file.image import Image
from zope.app.file.testing import AppFileLayer
from zope.app.file.tests import http
from zope.app.file.tests.test_image import zptlogo


class StringIO(BytesIO):

    def __init__(self, contents):
        if not isinstance(contents, bytes):
            contents = contents.encode("ascii")
        super().__init__(contents)


class BrowserTestCase(unittest.TestCase):

    layer = AppFileLayer

    def setUp(self):
        super().setUp()
        self._testapp = TestApp(self.layer.make_wsgi_app())

    def getRootFolder(self):
        return self.layer.getRootFolder()

    def commit(self):
        pass

    def checkForBrokenLinks(self, orig_response, path, basic=None):
        response = self.publish(path, basic=basic)
        try:
            links = response.html.find_all('a')
        except (AttributeError, TypeError):
            # Not html
            return

        for link in links:
            href = link.attrs['href']
            if href.endswith('@@SelectedManagementView.html'):
                # We don't install this at test time
                continue

            if not href.startswith('/'):
                href = path.rsplit('/', 1)[0] + '/' + href
            try:
                self.publish(href, basic=basic)
            except ComponentLookupError:
                # PrincipalSource, not installed at testing
                pass

    def publish(self, path, basic=None, form=None, headers=None):
        assert basic
        self._testapp.authorization = ('Basic', tuple(basic.split(':')))
        env = {'wsgi.handleErrors': False}
        if form:
            upload_files = []
            for k, v in list(form.items()):
                if isinstance(v, StringIO):
                    upload_files.append((k,
                                         getattr(v, 'filename', 'a file'),
                                         v.getvalue()))
                    del form[k]
            response = self._testapp.post(path, params=form,
                                          upload_files=upload_files,
                                          extra_environ=env, headers=headers)
        else:
            response = self._testapp.get(
                path, extra_environ=env, headers=headers)

        response.getBody = lambda: response.unicode_normal_body
        response.getStatus = lambda: response.status_int
        response.getHeader = lambda n: response.headers[n]
        return response


class FileTest(BrowserTestCase):

    content = b'File <Data>'

    def addFile(self):
        file = File(self.content)
        root = self.getRootFolder()
        root['file'] = file
        self.commit()

    def testAddForm(self):
        response = self.publish(
            '/+/zope.app.file.File=',
            basic='mgr:mgrpw')
        self.assertEqual(response.getStatus(), 200)
        body = response.getBody()
        self.assertIn('Add a File', body)
        self.assertIn('Content Type', body)
        self.assertIn('Data', body)
        self.assertIn('Object Name', body)
        self.assertIn('"Add"', body)
        self.checkForBrokenLinks(response, '/+/zope.app.file.File=',
                                 'mgr:mgrpw')

    def testAdd(self):
        response = self.publish(
            '/+/zope.app.file.File=',
            form={'type_name': 'zope.app.file.File',
                  'field.data': StringIO('A file'),
                  'field.data.used': '',
                  'add_input_name': 'file',
                  'UPDATE_SUBMIT': 'Add'},
            basic='mgr:mgrpw')
        self.assertEqual(response.getStatus(), 302)
        self.assertEqual(response.getHeader('Location'),
                         'http://localhost/@@contents.html')
        root = self.getRootFolder()
        self.assertIn('file', root)
        file = root['file']
        self.assertEqual(file.data, b'A file')

    def testAddWithoutName(self):
        data = StringIO('File Contents')
        data.filename = "test.txt"
        response = self.publish(
            '/+/zope.app.file.File=',
            form={'type_name': 'zope.app.file.File',
                  'field.data': data,
                  'field.data.used': '',
                  'UPDATE_SUBMIT': 'Add'},
            basic='mgr:mgrpw')
        self.assertEqual(response.getStatus(), 302)
        self.assertEqual(response.getHeader('Location'),
                         'http://localhost/@@contents.html')
        root = self.getRootFolder()
        self.assertIn('test.txt', root, list(root))
        file = root['test.txt']
        self.assertEqual(file.data, b'File Contents')

    def testEditForm(self):
        self.addFile()
        response = self.publish(
            '/file/@@edit.html',
            basic='mgr:mgrpw')
        self.assertEqual(response.getStatus(), 200)
        body = response.getBody()
        self.assertIn('Change a file', body)
        self.assertIn('Content Type', body)
        self.assertIn('Data', body)
        self.assertIn(escape(self.content.decode('ascii')), body)
        self.checkForBrokenLinks(response, '/file/@@edit.html', 'mgr:mgrpw')

    def testEdit(self):
        self.addFile()
        response = self.publish(
            '/file/@@edit.html',
            form={'field.data': '<h1>A File</h1>',
                  'field.data.used': '',
                  'field.contentType': 'text/plain',
                  'UPDATE_SUBMIT': 'Edit'},
            basic='mgr:mgrpw')
        self.assertEqual(response.getStatus(), 200)
        body = response.getBody()
        self.assertIn('Change a file', body)
        self.assertIn('Content Type', body)
        self.assertIn('Data', body)
        self.assertIn(escape('<h1>A File</h1>'), body)
        root = self.getRootFolder()
        file = root['file']
        self.assertEqual(file.data, b'<h1>A File</h1>')
        self.assertEqual(file.contentType, 'text/plain')

    def testUploadForm(self):
        self.addFile()
        response = self.publish(
            '/file/@@upload.html',
            basic='mgr:mgrpw')
        self.assertEqual(response.getStatus(), 200)
        body = response.getBody()
        self.assertIn('Upload a file', body)
        self.assertIn('Content Type', body)
        self.assertIn('Data', body)
        self.assertNotIn(escape(self.content.decode('ascii')), body)
        self.checkForBrokenLinks(response, '/file/@@upload.html', 'mgr:mgrpw')

    def testUpload(self):
        self.addFile()
        response = self.publish(
            '/file/@@upload.html',
            form={'field.data': StringIO('<h1>A file</h1>'),
                  'field.data.used': '',
                  'field.contentType': 'text/plain',
                  'UPDATE_SUBMIT': 'Change'},
            basic='mgr:mgrpw')
        self.assertEqual(response.getStatus(), 200)
        body = response.getBody()
        self.assertIn('Upload a file', body)
        self.assertIn('Content Type', body)
        self.assertIn('Data', body)
        self.assertNotIn(escape('<h1>A File</h1>'), body)
        root = self.getRootFolder()
        file = root['file']
        self.assertEqual(file.data, b'<h1>A file</h1>')
        self.assertEqual(file.contentType, 'text/plain')

    def testIndex(self):
        self.addFile()
        response = self.publish(
            '/file/@@index.html',
            basic='mgr:mgrpw')
        self.assertEqual(response.getStatus(), 200)
        body = response.body
        self.assertEqual(body, self.content)
        self.checkForBrokenLinks(response, '/file/@@index.html', 'mgr:mgrpw')

    def testPreview(self):
        self.addFile()
        response = self.publish(
            '/file/@@preview.html',
            basic='mgr:mgrpw')
        self.assertEqual(response.getStatus(), 200)
        body = response.getBody()
        self.assertIn('<iframe src="."', body)
        self.checkForBrokenLinks(response, '/file/@@preview.html', 'mgr:mgrpw')


class ImageTest(BrowserTestCase):

    content = zptlogo

    def addImage(self):
        image = Image(self.content)
        root = self.getRootFolder()
        root['image'] = image
        self.commit()

    def testAddForm(self):
        response = self.publish(
            '/+/zope.app.file.Image=',
            basic='mgr:mgrpw')
        self.assertEqual(response.getStatus(), 200)
        body = response.getBody()
        self.assertIn('Add an Image', body)
        self.assertIn('Content Type', body)
        self.assertIn('Data', body)
        self.assertIn('Object Name', body)
        self.assertIn('"Add"', body)
        self.checkForBrokenLinks(response, '/+/zope.app.file.Image=',
                                 'mgr:mgrpw')

    def testAdd(self):
        response = self.publish(
            '/+/zope.app.file.Image=',
            form={'type_name': 'zope.app.image.Image',
                  'field.data': StringIO(self.content),
                  'field.data.used': '',
                  'add_input_name': 'image',
                  'UPDATE_SUBMIT': 'Add'},
            basic='mgr:mgrpw')
        self.assertEqual(response.getStatus(), 302)
        self.assertEqual(response.getHeader('Location'),
                         'http://localhost/@@contents.html')
        root = self.getRootFolder()
        self.assertIn('image', root)
        image = root['image']
        self.assertEqual(image.data, self.content)

    def testAddWithoutName(self):
        data = StringIO(self.content)
        data.filename = "test.gif"
        response = self.publish(
            '/+/zope.app.file.Image=',
            form={'type_name': 'zope.app.image.Image',
                  'field.data': data,
                  'field.data.used': '',
                  'UPDATE_SUBMIT': 'Add'},
            basic='mgr:mgrpw')
        self.assertEqual(response.getStatus(), 302)
        self.assertEqual(response.getHeader('Location'),
                         'http://localhost/@@contents.html')
        root = self.getRootFolder()
        self.assertIn('test.gif', root, list(root))
        image = root['test.gif']
        self.assertEqual(image.data, self.content)

    def testUploadForm(self):
        self.addImage()
        response = self.publish(
            '/image/@@upload.html',
            basic='mgr:mgrpw')
        self.assertEqual(response.getStatus(), 200)
        body = response.getBody()
        self.assertIn('Upload an image', body)
        self.assertIn('Content Type', body)
        self.assertIn('Data', body)
        self.assertIn('1 KB 16x16', body)
        self.checkForBrokenLinks(response, '/image/@@upload.html', 'mgr:mgrpw')

    def testUpload(self):
        self.addImage()
        response = self.publish(
            '/image/@@upload.html',
            form={'field.data': StringIO(''),
                  'field.data.used': '',
                  'UPDATE_SUBMIT': 'Change'},
            basic='mgr:mgrpw')
        self.assertEqual(response.getStatus(), 200)
        body = response.getBody()
        self.assertIn('Upload an image', body)
        self.assertIn('Content Type', body)
        self.assertIn('Data', body)
        self.assertIn('0 KB ?x?', body)
        root = self.getRootFolder()
        image = root['image']
        self.assertEqual(image.data, b'')
        self.assertEqual(image.contentType, 'image/gif')

    def testUpload_only_change_content_type(self):
        self.addImage()
        response = self.publish(
            '/image/@@upload.html',
            form={'field.contentType': 'image/png',
                  'UPDATE_SUBMIT': 'Change'},
            basic='mgr:mgrpw')
        self.assertEqual(response.getStatus(), 200)
        body = response.getBody()
        self.assertIn('Upload an image', body)
        self.assertIn('Content Type', body)
        self.assertIn('Data', body)
        self.assertIn('1 KB 16x16', body)
        root = self.getRootFolder()
        image = root['image']
        self.assertEqual(image.data, self.content)
        self.assertEqual(image.contentType, 'image/png')

    def testIndex(self):
        self.addImage()
        response = self.publish(
            '/image/@@index.html',
            basic='mgr:mgrpw')
        self.assertEqual(response.getStatus(), 200)
        body = response.body
        self.assertEqual(body, self.content)
        self.checkForBrokenLinks(response, '/image/@@index.html', 'mgr:mgrpw')

    def testPreview(self):
        self.addImage()
        response = self.publish(
            '/image/@@preview.html',
            basic='mgr:mgrpw')
        self.assertEqual(response.getStatus(), 200)
        body = response.getBody()
        self.assertIn('<iframe src="."', body)
        self.checkForBrokenLinks(
            response, '/image/@@preview.html', 'mgr:mgrpw')


def test_suite():
    def _make_doctest(fname):
        test = doctest.DocFileSuite(
            fname,
            globs={'http': http,
                   'encodeMultipartFormdata': encodeMultipartFormdata},
            optionflags=(doctest.ELLIPSIS
                         | doctest.NORMALIZE_WHITESPACE
                         | doctest.IGNORE_EXCEPTION_DETAIL))
        test.layer = AppFileLayer
        return test

    url = _make_doctest('../url.rst')
    file = _make_doctest('../file.rst')
    return unittest.TestSuite((
        unittest.defaultTestLoader.loadTestsFromName(__name__),
        url,
        file,
    ))



##############################################################################
#
# Copyright (c) 2001, 2002 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Define view component for naive file editing.

"""
__docformat__ = 'restructuredtext'

from zope.size.interfaces import ISized
from zope.traversing.browser.absoluteurl import absoluteURL

from zope.app.file.browser.file import FileView
from zope.app.file.browser.file import cleanupFileName


class ImageData(FileView):

    def __call__(self):
        return self.show()

    def tag(self, height=None, width=None, alt=None,
            scale=0, xscale=0, yscale=0, css_class=None, **args):
        """
        Generate an HTML IMG tag for this image, with customization.
        Arguments to ``self.tag()`` can be any valid attributes of an IMG tag.
        `src` will always be an absolute pathname, to prevent redundant
        downloading of images. Defaults are applied intelligently for
        `height`, `width`, and `alt`. If specified, the `scale`, `xscale`,
        and `yscale` keyword arguments will be used to automatically adjust
        the output height and width values of the image tag.

        Since 'class' is a Python reserved word, it cannot be passed in
        directly in keyword arguments which is a problem if you are
        trying to use ``tag()`` to include a CSS class. The `tag()` method
        will accept a `css_class` argument that will be converted to
        ``class`` in the output tag to work around this.
        """
        if width is None:
            width = self.context.getImageSize()[0]
        if height is None:
            height = self.context.getImageSize()[1]

        # Auto-scaling support
        xdelta = xscale or scale
        ydelta = yscale or scale

        if xdelta and width:
            width = str(int(round(int(width) * xdelta)))
        if ydelta and height:
            height = str(int(round(int(height) * ydelta)))

        result = '<img '
        if self.request is not None:
            result = '<img src="%s"' % absoluteURL(self.context, self.request)

        if alt is None:
            alt = getattr(self, 'title', '')
        result = f'{result} alt="{alt}"'

        if height is not None:
            result = f'{result} height="{height}"'

        if width is not None:
            result = f'{result} width="{width}"'

        if 'border' not in [a.lower() for a in args]:
            result = '%s border="0"' % result

        if css_class is not None:
            result = f'{result} class="{css_class}"'

        for key, value in args.items():
            result = f'{result} {key}="{value}"'

        return '%s />' % result


class ImageUpload:
    """Image edit view mix-in that provides access to image size info"""

    def size(self):
        sized = ISized(self.context)
        return sized.sizeForDisplay()


class ImageAdd:

    def update(self):

        if self.update_status is not None:
            # We've been called before. Just return the previous result.
            return self.update_status

        # set the name of the image if not defined
        if not self.request.form.get('add_input_name'):
            filename = getattr(self.request.form.get("field.data"),
                               "filename", None)
            if filename is not None:
                self.request.form["add_input_name"] = cleanupFileName(filename)

        return super().update()



##############################################################################
#
# Copyright (c) 2006 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
# This package is developed by the Zope Toolkit project, documented here:
# http://docs.zope.org/zopetoolkit
# When developing and releasing this package, please follow the documented
# Zope Toolkit policies as described by this documentation.
##############################################################################
"""Setup for zope.app.file package

"""
import os

from setuptools import find_packages
from setuptools import setup


def read(*rnames):
    with open(os.path.join(os.path.dirname(__file__), *rnames)) as f:
        return f.read()


setup(name='zope.app.file',
      version='5.0',
      author='Zope Foundation and Contributors',
      author_email='zope-dev@zope.org',
      description='File and Image -- Zope 3 Content Components',
      long_description=(
          read('README.rst')
          + '\n\n.. contents::\n\n' +
          read('src', 'zope', 'app', 'file', 'browser', 'file.rst')
          + '\n\n' +
          read('src', 'zope', 'app', 'file', 'browser', 'url.rst')
          + '\n\n' +
          read('CHANGES.rst')
      ),
      keywords="zope3 file image content",
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Environment :: Web Environment',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: Zope Public License',
          'Programming Language :: Python',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.8',
          'Programming Language :: Python :: 3.9',
          'Programming Language :: Python :: 3.10',
          'Programming Language :: Python :: 3.11',
          'Programming Language :: Python :: 3.12',
          'Programming Language :: Python :: 3.13',
          'Programming Language :: Python :: Implementation :: CPython',
          'Programming Language :: Python :: Implementation :: PyPy',
          'Natural Language :: English',
          'Operating System :: OS Independent',
          'Topic :: Internet :: WWW/HTTP',
          'Framework :: Zope :: 3',
      ],
      url='https://github.com/zopefoundation/zope.app.file',
      license='ZPL 2.1',
      packages=find_packages('src'),
      package_dir={'': 'src'},
      namespace_packages=['zope', 'zope.app'],
      python_requires='>=3.8',
      extras_require={
          'test': [
              'webtest',
              'zope.app.appsetup >= 4.0.0',
              'zope.app.basicskin >= 4.0.0',
              'zope.app.exception >= 4.0.1',
              'zope.app.folder >= 4.0.0',
              'zope.app.http >= 4.0.1',
              'zope.app.rotterdam >= 4.0.0',
              'zope.app.securitypolicy',
              'zope.app.wsgi >= 5.3',
              'zope.dublincore',
              'zope.login',
              'zope.principalannotation',
              'zope.publisher >= 3.12',
              'zope.testing',
              'zope.testrunner',
          ]
      },
      install_requires=[
          'setuptools',
          'transaction',
          'persistent',
          'zope.app.content >= 4.0.0',
          'zope.app.form >= 5.0.0',
          'zope.app.publication',
          'zope.contenttype >= 4.0.0',
          'zope.datetime',
          'zope.dublincore >= 4.0.0',
          'zope.event',
          'zope.exceptions',
          'zope.filerepresentation',
          'zope.i18nmessageid >= 4.1.0',
          'zope.interface',
          'zope.schema',
          'zope.site',
          'zope.size',
      ],
      include_package_data=True,
      zip_safe=False,
      )



##############################################################################
#
# Copyright (c) 2001, 2002 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Image content type implementation

"""
__docformat__ = 'restructuredtext'

import struct
from io import BytesIO

from zope.contenttype import guess_content_type
from zope.interface import implementer
from zope.size import byteDisplay
from zope.size.interfaces import ISized

from zope.app.file.file import File
from zope.app.file.i18n import ZopeMessageFactory as _
from zope.app.file.interfaces import IImage


@implementer(IImage)
class Image(File):

    def __init__(self, data=b''):
        '''See interface `IFile`'''
        contentType, self._width, self._height = getImageInfo(data)
        super().__init__(data, contentType)

    def _setData(self, data):
        super()._setData(data)

        contentType, self._width, self._height = getImageInfo(self._data)
        if contentType:
            self.contentType = contentType

    def getImageSize(self):
        '''See interface `IImage`'''
        return (self._width, self._height)

    data = property(File._getData, _setData)


@implementer(ISized)
class ImageSized:

    def __init__(self, image):
        self._image = image

    def sizeForSorting(self):
        '''See `ISized`'''
        return ('byte', self._image.getSize())

    def sizeForDisplay(self):
        '''See `ISized`'''
        w, h = self._image.getImageSize()
        if w < 0:
            w = '?'
        if h < 0:
            h = '?'
        bytes = self._image.getSize()
        byte_size = byteDisplay(bytes)
        mapping = byte_size.mapping
        if mapping is None:
            mapping = {}
        mapping.update({'width': str(w), 'height': str(h)})
        # TODO the way this message id is defined, it won't be picked up by
        # i18nextract and never show up in message catalogs
        return _(byte_size + ' ${width}x${height}', mapping=mapping)


class FileFactory:

    def __init__(self, context):
        self.context = context

    def __call__(self, name, content_type, data):
        if not content_type and data:
            content_type, _width, _height = getImageInfo(data)
        if not content_type:
            content_type, _encoding = guess_content_type(name, data, '')

        if content_type.startswith('image/'):
            return Image(data)

        return File(data, content_type)


def getImageInfo(data):
    data = bytes(data)
    size = len(data)
    height = -1
    width = -1
    content_type = ''

    # handle GIFs
    if (size >= 10) and data[:6] in (b'GIF87a', b'GIF89a'):
        # Check to see if content_type is correct
        content_type = 'image/gif'
        w, h = struct.unpack(b"<HH", data[6:10])
        width = int(w)
        height = int(h)

    # See PNG 2. Edition spec (http://www.w3.org/TR/PNG/)
    # Bytes 0-7 are below, 4-byte chunk length, then 'IHDR'
    # and finally the 4-byte width, height
    elif ((size >= 24) and data.startswith(b'\211PNG\r\n\032\n')
          and (data[12:16] == b'IHDR')):
        content_type = 'image/png'
        w, h = struct.unpack(b">LL", data[16:24])
        width = int(w)
        height = int(h)

    # Maybe this is for an older PNG version.
    elif ((size >= 16)
          and data.startswith(b'\211PNG\r\n\032\n')):  # pragma: no cover
        # Check to see if we have the right content type
        content_type = 'image/png'
        w, h = struct.unpack(">LL", data[8:16])
        width = int(w)
        height = int(h)

    # handle JPEGs
    elif (size >= 2) and data.startswith(b'\xff\xd8'):
        content_type = 'image/jpeg'
        jpeg = BytesIO(data)
        jpeg.read(2)
        b = jpeg.read(1)
        try:
            w = -1
            h = -1
            while b and ord(b) != 0xDA:
                while ord(b) != 0xFF:
                    b = jpeg.read(1)
                while ord(b) == 0xFF:
                    b = jpeg.read(1)
                if 0xC0 <= ord(b) <= 0xC3:
                    jpeg.read(3)
                    h, w = struct.unpack(b">HH", jpeg.read(4))
                    break
                else:
                    jpeg.read(int(struct.unpack(b">H", jpeg.read(2))[0]) - 2)
                b = jpeg.read(1)
            width = int(w)
            height = int(h)
        except (struct.error, ValueError):  # pragma: no cover
            pass

    # handle BMPs
    elif (size >= 30) and data.startswith(b'BM'):
        kind = struct.unpack(b"<H", data[14:16])[0]
        if kind == 40:  # Windows 3.x bitmap
            content_type = 'image/x-ms-bmp'
            width, height = struct.unpack(b"<LL", data[18:26])

    return content_type, width, height
